from .iring import * 
from morebs2.numerical_generator import prg_choose_n
from morebs2.dep_ord import * 
from copy import deepcopy 

def index_in_OOC(ooc,element): 
    for (i,c) in enumerate(ooc): 
        if element in c: 
            return i 
    return -1

class IsoRingedChain:

    def __init__(self,ir_list,prng=None): 
        for ir in ir_list: assert type(ir) == IsoRing
        
        d = dict() 
        for q_ in ir_list: 
            d[q_.idn_tag()] = (q_.dc_set(True),q_.dc_set(False)) 

        # check for valid order of cracking 
        ooc,stat = calculate_dependency_order(d)
        assert stat 

        self.ir_dict = {ir.idn_tag(): ir for ir in ir_list}
        self.ooc = ooc 

        # information used in defensive procedure against Cracker 
        if type(prng) == type(None): 
            self.prng = default_std_Python_prng(output_range=[-10000,10000],rounding_depth=0) 
        else: 
            self.prng = prng 

        # set of IsoRing identifiers 
        self.finished_targetset = set()
        self.current_ir_target = set()  
        return

    #------------------- methods to interact with <Cracker>

    def set_current_ir_targetset(self,targetset):
        assert type(targetset) == set 
        if not self.accept_cracker_targetset(targetset): 
            return False 
        self.current_ir_target = targetset 
        return True 

    """
    Cracker targets instance in units of <IsoRing> sets. 
    Method uses method<accept_cracker_target>. 
    """
    def accept_cracker_targetset(self,targetset): 
        for t in targetset: 
            if not self.accept_cracker_target(t): return False 
        return True 

    """
    responds with boolean if Cracker can proceed with targeting 
    <IsoRing> `target_ir_idn`, based on 
    """
    def accept_cracker_target(self,target_ir_idn): 
        ir = self.fetch_IsoRing(target_ir_idn)  
        if type(ir) == type(None): return False 

        dep = ir.dc_set(True) 
        if dep.intersection(self.finished_targetset) != dep: 
            return False 
        return True 

    #------------------------------------------------------- 

    def repr_dict_for_IsoRings(self,isoring_idn): 
        D = {} 
        for i in isoring_idn: 
            I = self.fetch_IsoRing(i)
            assert type(I) != type(None) 
            D[I.idn_tag()] = I.current_sec_index
        return D 
 
    def register_cracked_IsoRings(self,wanted_finishes:set,recracks:set): 
        self.finished_targetset |= wanted_finishes

        Q = wanted_finishes | recracks 
        for q in Q: 
            self.shift_IsoRing(q)  
        return 

    def shift_IsoRing(self,ir_idn): 
        I = self.fetch_IsoRing(ir_idn)
        return I.register_cracked_sec_index(self.prng)

    #-------------------------------------------------------- 

    def fetch_IsoRing(self,ir_idn):
        if ir_idn not in self.ir_dict: return None 
        return self.ir_dict[ir_idn] 

    @staticmethod
    def prng__add_depANDcodep_to_IsoRingList(ir_list,prng,codep_ratio=0.0):
        assert len(ir_list) > 0 
        assert 0.0 <= codep_ratio <= 1.0 

        idns = [] 
        ir_dict = {} 
        for ir in ir_list:
            idns.append(ir.idn_tag()) 
            ir.clear_depANDcodep_sets()
            ir_dict[ir.idn_tag()] = ir
        
        oodc = IsoRingedChain.prng__idns_to_order_of_depANDcodep(idns,prng,codep_ratio)
        for j in range(len(oodc)): 
            depset = set() 
            for i in range(0,j): 
                depset |= oodc[i] 

            # add codep and dep for each in the set 
            codeps = oodc[j]
            for j_ in codeps: 
                cds = codeps - {j_} 
                ir = ir_dict[j_] 
                ir.assign_DC_set(depset,cds) 
    
    @staticmethod
    def prng__idns_to_order_of_depANDcodep(idns,prng,codep_ratio):
        assert len(set(idns)) == len(idns)

        def prg_(): 
            return int(prng())

        total_conn = len(idns) - 1 
        codep_conn = int(ceil(total_conn * codep_ratio))
        L = prg_seqsort(idns,prg_) 
        L_ = []

        while codep_conn > 0 and len(L) > 0: 
            x = modulo_in_range(prg_(),[1,codep_conn+1]) 
            S = set() 
            i = prg_() % len(L)
            l = L.pop(i)
            S |= {l}

            for _ in range(x):
                if len(L) == 0: break  
                i = prg_() % len(L)
                l = L.pop(i)
                S |= {l} 
            
            L_.append(S) 
            codep_conn -= x 

        while len(L) > 0: 
            L_.append({L.pop(0)})

        L_ = prg_seqsort(L_,prg_)
        return L_

    @staticmethod 
    def list_of_vectors_to_IsoRingedChain(vec_list,prng,num_blooms_range=[DEFAULT_NUM_BLOOMS,DEFAULT_NUM_BLOOMS+1],
        ratio_of_feedback_functions_type_1:float=1.0,codep_ratio=0.0,shuffle_isorepr:bool=True):  

        def prg_(): return int(prng()) 

        # shuffle `vec_list`
        vec_list = prg_seqsort(vec_list,prg_)

        # generate the list of <Sec> instances 
        sec_list = [] 
        for (i,vec) in enumerate(vec_list): 
            num_optima = modulo_in_range(prg_(),DEFAULT_NUM_OPTIMA_RANGE)
            set_actual_as_max_pr = bool(prg_() % 2)
            sec = Sec.vec_to_bare_instance(vec,singleton_distance=DEFAULT_SINGLETON_DISTANCE_RANGE,\
                num_optima=num_optima,prng=prng,idn_tag=i,set_actual_as_max_pr=set_actual_as_max_pr)
            sec_list.append(sec) 

        # get the <Sec> indexes with feedback function type 1
        num_sec_ftype_1 = int(ceil(ratio_of_feedback_functions_type_1 * len(sec_list))) 
        index_list = [_ for _ in range(len(sec_list))] 

        ftype_1_indices = prg_choose_n(index_list,num_sec_ftype_1,prg_,is_unique_picker=True)

        ir_list = [] 
        for (i,sec) in enumerate(sec_list): 
            # transform each <Sec> into an <IsoRing> 
            feedback_function_type = 1 if i in ftype_1_indices else 0 
            ir = IsoRing.generate_IsoRing_from_one_secret(sec,prng,feedback_function_type,\
                num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
                sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,\
                optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE)
            ir_list.append(ir)

        IsoRingedChain.prng__add_depANDcodep_to_IsoRingList(ir_list,prng,codep_ratio)

        if shuffle_isorepr: 
            for ir in ir_list: 
                ir.cracked_sec_indices.append(ir.actual_sec_index)
                ir.switch_iso_repr(prng)
                ir.cracked_sec_indices.pop(-1) 

        irc = IsoRingedChain(ir_list) 
        return irc 


from isoring.secrets.bloominstein import * 
import unittest

### lone file test 
"""
python3 -m tests.test_bloominstein 
"""
###
class BloomOfSecretClass(unittest.TestCase):

    def test__BloomOfSecret__next__case1(self):
        singleton_range = [-100,100] 
        dimension = 5 
        num_optima = 5 
        prng = default_std_Python_prng(integer_seed=55,output_range=[0,1000],rounding_depth=3)
        s = Sec.generate_bare_instance(singleton_range,dimension,num_optima,prng,idn_tag=0,set_actual_as_max_pr=True)

        XX = BloomOfSecret(s,prng,num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
                sec_vec_multiplier_range=[0.5,2.0],optima_multiplier_range=[0.5,2.0])

        for i in range(DEFAULT_NUM_BLOOMS):
            assert type(next(XX)) != type(None) 

        assert type(next(XX)) == type(None) 
        return

    def test__BloomOfSecret__next__case2(self):
        singleton_range = [-100,100] 
        dimension = 5 
        num_optima = 5 
        prng = default_std_Python_prng(integer_seed=55,output_range=[0,1000],rounding_depth=3)
        s = Sec.generate_bare_instance(singleton_range,dimension,num_optima,prng,idn_tag=0,set_actual_as_max_pr=True)

        num_blooms = DEFAULT_BLOOM_VECTOR_DIM_RANGE[1] - DEFAULT_BLOOM_VECTOR_DIM_RANGE[0]

        try: 
            XX = BloomOfSecret(s,prng,num_blooms=num_blooms,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
                sec_vec_multiplier_range=[0.5,2.0],optima_multiplier_range=[0.5,2.0])
            assert False 
        except: 
            assert True 

if __name__ == '__main__':
    unittest.main()



from isoring.brute_forcer.brute_force_env import * 
import unittest

def veclistANDprng_sample_U(): 
    prng = prg__LCG(45.43,543.21,-42315.6666,6797.0)
    vec_size = [3,5,4,7] 
    vec_list = [] 
    for i in range(16): 
        V = one_vec(prng,vec_size[i % 4],[-2000,2000])
        vec_list.append(V) 
    return vec_list,prng 

### lone file test 
"""
python3 -m tests.test_brute_force_env  
"""
###
class BruteForceEnvClass(unittest.TestCase):

    '''
    successful crack, completely wrong answer
    '''
    def test__BruteForceEnv__next__case1(self):
        vec_list,prng = veclistANDprng_sample_U() 
        prng2 = prg__LCG(-43,-45,61,-7171)
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list[:3],prng2,\
            num_blooms_range=[DEFAULT_NUM_BLOOMS-2,DEFAULT_NUM_BLOOMS+1],ratio_of_feedback_functions_type_1=0.0,\
            codep_ratio=0.9,shuffle_isorepr=False)
        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,[0.0,0.0],[1.0,1.0],\
                [1.0,1.0],0.0,0.0,0.0)

        print("\t\tTESTING BRUTE FORCE ENV #1") 
        crck = Cracker(bi,10,verbose=True) 
        bfe = BruteForceEnv(crck,irc,prng,True)

        assert not crck.halted 
        while not bfe.is_finished():
            next(bfe) 
        assert bfe.num_iter == 316, "got {}".format(bfe.num_iter)
        assert len(crck.csoln) == 3 

        for i in range(3): 
            soln = crck.soln_for_IsoRing(i)
            ir = irc.fetch_IsoRing(i) 
            V = ir.actual_sec_vec()
            assert not np.any(soln[1] == V) 
        print("-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/")

    '''
    cracking process terminated midway, no energy left
    '''
    def test__BruteForceEnv__next__case2(self): 
        print("\t\tTESTING BRUTE FORCE ENV #2") 
        vec_list,prng = veclistANDprng_sample_U() 
        prng2 = prg__LCG(-43,-45,61,-7171)
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list[:3],prng2,\
            num_blooms_range=[DEFAULT_NUM_BLOOMS-2,DEFAULT_NUM_BLOOMS+1],ratio_of_feedback_functions_type_1=0.0,\
            codep_ratio=0.9,shuffle_isorepr=False)
        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,[1.0,1.0],[1.0,1.0],\
                [1.0,1.0],0.0,0.0,0.0)

        crck = Cracker(bi,10,100,True)  
        bfe = BruteForceEnv(crck,irc,prng,True) 

        assert not crck.halted 
        while not bfe.is_finished(): 
            next(bfe) 
        assert crck.halted
        assert len(crck.csoln) == 2, "got {}".format(crck.csoln.D)

        soln = crck.soln_for_IsoRing(2)
        ir = irc.fetch_IsoRing(2) 
        V = ir.actual_sec_vec()
        assert np.all(soln[1] == V)
        assert bfe.num_iter == 65, "got {}".format(bfe.num_iter)

        print("-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/")
        return

    '''
    cracking process on two IsoRings, wrong optima  
    '''
    def test__BruteForceEnv__next__case3(self): 
        print("\t\tTESTING BRUTE FORCE ENV #3") 
        vec_list,prng = veclistANDprng_sample_U() 

        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list,prng,\
            num_blooms_range=[DEFAULT_NUM_BLOOMS-2,DEFAULT_NUM_BLOOMS+1],ratio_of_feedback_functions_type_1=0.0,\
            codep_ratio=0.0,shuffle_isorepr=False)

        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,[0.0,0.0],[1.0,1.0],\
                [1.0,1.0],1.0,0.0,0.0)

        crck = Cracker(bi,10,verbose=True) 

        bfe = BruteForceEnv(crck,irc,prng,True) 

        for i in range(50000): 
            next(bfe) 

        Q = set(crck.csoln.D.keys())


        Q2 = set()
        for x in range(6): 
            Q2 |= bi.order_of_cracking[x] 
        assert Q == Q2 

        for k in Q2: 
            soln = crck.soln_for_IsoRing(k)
            ir = irc.fetch_IsoRing(k) 
            V = ir.actual_sec_vec()
            assert not np.any(soln[1] == V)

            # correct Sec index 
            sec_index = list(crck.csoln.D[k].keys()).pop() 
            assert sec_index == ir.actual_sec_index

            # not the correct optimum 
            S = ir.fetch_Sec(sec_index)
            assert S.seq_index() != soln[0] 
            assert vector_to_string(soln[1],float) in S.opm 

        print("-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/")

    """
    cracking on a co-dependent set of size 3; recracking required. 
    """
    def test__BruteForceEnv__next__case4(self):
        print("\t\tTESTING BRUTE FORCE ENV #4") 
        vec_list,prng = veclistANDprng_sample_U() 

        prng2 = prg__LCG(-43,-45,61,-7171)
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list,prng2,\
            num_blooms_range=[DEFAULT_NUM_BLOOMS-2,DEFAULT_NUM_BLOOMS+1],ratio_of_feedback_functions_type_1=0.0,\
            codep_ratio=0.9,shuffle_isorepr=False)
        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,[0.0,0.0],[1.0,1.0],\
                [1.0,1.0],1.0,0.0,1.0)

        crck = Cracker(bi,10,verbose=True) 
        bfe = BruteForceEnv(crck,irc,prng,True) 

        assert not crck.halted 

        for i in range(20000):  
            next(bfe) 

        assert len(crck.csoln) == 4
        assert len(bfe.cbridges) == 2 == len(crck.active_cracklings)
        assert len(crck.csoln.D[0]) == len(crck.csoln.D[14]) == 1
        assert len(crck.csoln.D[15]) == 2

        print("-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/")


    """
    correct BackgroundInfo for Cracker; Cracker does not have enough energy to completely crack 
    IsoRingedChain. 
    """
    def test__BruteForceEnv__next__case5(self):
        print("\t\tTESTING BRUTE FORCE ENV #5") 
        vec_list,prng = veclistANDprng_sample_U() 

        prng2 = prg__LCG(-43,-45,61,-7171)
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list[:3],prng2,\
            num_blooms_range=[DEFAULT_NUM_BLOOMS-2,DEFAULT_NUM_BLOOMS+1],ratio_of_feedback_functions_type_1=0.0,\
            codep_ratio=0.9,shuffle_isorepr=False) 
        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,[1.0,1.0],[1.0,1.0],\
                [1.0,1.0],0.0,0.0,0.0)

        crck = Cracker(bi,10,100)  
        bfe = BruteForceEnv(crck,irc,prng,True) 

        assert not crck.halted 
        while not bfe.is_finished(): 
            next(bfe) 
         
        assert len(crck.csoln) == 2,"got {}".format(crck.csoln.D)

        soln = crck.soln_for_IsoRing(2)
        ir = irc.fetch_IsoRing(2) 
        V = ir.actual_sec_vec()
        assert np.all(soln[1] == V)
        assert bfe.num_iter == 65, "got {}".format(bfe.num_iter)

        print("-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/")


    """
    completely wrong 
    """
    def test__BruteForceEnv__next__case6(self):
        print("\t\tTESTING BRUTE FORCE ENV #6") 
        vec_list,prng = veclistANDprng_sample_U() 

        prng2 = prg__LCG(-43,-45,61,-7171)
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list[:5],prng2,\
            num_blooms_range=[DEFAULT_NUM_BLOOMS-2,DEFAULT_NUM_BLOOMS+1],ratio_of_feedback_functions_type_1=0.0,\
            codep_ratio=0.9,shuffle_isorepr=False)
        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,[1.0,1.0],[1.0,1.0],\
                [0.0,0.0],0.0,0.0,0.0)

        crck = Cracker(bi,10,50000)  
        bfe = BruteForceEnv(crck,irc,prng,True) 

        assert not crck.halted 
        while not bfe.is_finished(): 
            next(bfe) 
        assert len(crck.csoln) == 0 

        print("-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/")


if __name__ == '__main__':
    unittest.main()



from ..brute_forcer.brute_force_env import * 
from morebs2.message_streamer import * 

DEFAULT_MAX_VECTOR_LENGTH = 5 
DEFAULT_MAX_NUM_VECTORS = 50 

def read_vector_file(fp,max_vector_length=DEFAULT_MAX_VECTOR_LENGTH,max_num_vectors=DEFAULT_MAX_NUM_VECTORS): 
    
    vec_list = [] 
    with open(fp, 'r') as fin:
        for line in fin:
            if max_num_vectors == 0: break 

            row = line.strip()
            vec = string_to_vector(row,float)
            assert 0 < len(vec) <= max_vector_length
            vec_list.append(vec) 
            max_num_vectors -= 1 
    
    return vec_list 
    
def load_vector_file_into_IsoRingedChain(fp,prng=default_std_Python_prng()):
    prg_ = prng_to_decimal_output(prng)
    codep_ratio = prg_() 

    vec_list = read_vector_file(fp) 
    
    return IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list,prng=default_std_Python_prng(),\
        num_blooms_range=[DEFAULT_NUM_BLOOMS-1,DEFAULT_NUM_BLOOMS],\
        ratio_of_feedback_functions_type_1=1.0,codep_ratio=codep_ratio,shuffle_isorepr=True)

def simulation_default_BackgroundInfo_for_IsoRingedChain(irc,prng,allow_inaccuracies:bool,\
    allow_incomplete_info:bool,allow_wrong_OOC:bool,allow_one_shot_kill:bool,max_hop_size=9): 

    assert type(max_hop_size) in {int,np.int32,np.int64} and 2 <= max_hop_size <=9

    prg_ = prng_to_decimal_output(prng) 

    prioritize_actual_Sec_ratio = 1.0 
    if allow_inaccuracies: 
        actual_sec_vec_ratio_range = np.round(np.sort([prg_(),prg_()]),5)
        valid_bounds_ratio_range = np.round(np.sort([prg_(),prg_()]),5)
        suspected_isoring_to_sec_idn_error_ratio = prg_()
    else: 
        actual_sec_vec_ratio_range = [1.0,1.0]
        valid_bounds_ratio_range = [1.0,1.0] 
        suspected_isoring_to_sec_idn_error_ratio = 0.0 
         
    if allow_incomplete_info: 
        dim_covered_ratio_range = np.round(np.sort([prg_(),prg_()]),5)
    else: 
        dim_covered_ratio_range = [1.,1.]

    if allow_wrong_OOC: 
        shuffle_OOC_ratio = prg_() 
    else: 
        shuffle_OOC_ratio = 0.0 
        
    if allow_one_shot_kill:
        valid_one_shot_kill_ratio_range = [1.,1.]
    else: 
        valid_one_shot_kill_ratio_range = [0.,0.]

    bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng,actual_sec_vec_ratio_range,\
        dim_covered_ratio_range,valid_bounds_ratio_range,prioritize_actual_Sec_ratio,\
        shuffle_OOC_ratio,suspected_isoring_to_sec_idn_error_ratio,valid_one_shot_kill_ratio_range,[2,max_hop_size])
    return bi 

def instantiate_simulation_BruteForceEnv(bi,irc,cracker_energy:float): 
    prng = default_std_Python_prng() 
    crck = Cracker(bi,DEFAULT_MAX_NUM_VECTORS,cracker_energy,True)  
    bfe = BruteForceEnv(crck,irc,prng,True) 
    return bfe 

#-----------------------------------------------------------------------------------------------------------

def save_std_Python_random_state(fp): 
    # Save the current random generator state
    state = random.getstate()

    try: 
        # Optionally, save state to a file
        with open(fp, "wb") as f:
            pickle.dump(state, f)
        return True 
    except: 
        print("\t[!] invalid file for storing random state.")
        return False 

def load_std_Python_random_state(fp): 
    try: 
        with open(fp, "rb") as f:
            saved_state = pickle.load(f)
            random.setstate(saved_state)
        return True 
    except: 
        print("\t[!] invalid file for storing random state.")
        return False 


from .hypothesis import * 
from ..secrets.big_secret import * 


def prng_to_decimal_output(prng):

    def f(): 
        q1 = abs(prng()) 
        q2 = abs(prng())
        if q1 == 0 or q2 == 0: return 0.0 

        if q1 < q2: return q1 / q2
        return q2 / q1 
    return f 

# NOTE: this prng must exclusively output integers 
def n_swaps_on_seq(seq,n,prng):

    ix = [_ for _ in range(len(seq))] 
    ix = prg_seqsort(ix,prng) 

    for i in range(n): 
        # pop index 
        index = ix[i%len(ix)]
        x = seq.pop(index) 

        # re-insert 
        index = prng() % (len(seq) +1)

        seq.insert(index,x) 
    return seq 

class BackgroundInfo:

    """
    info := dict, <Isoring> identifier -> <Sec> index -> <HypStruct> 
    suspected_isoring_to_sec_idn := dict, <IsoRing> -> index of <Sec> most likely to be solution. 
    order_of_cracking := list, of sets of <IsoRing> identifiers, specifying the 
                         order that a <Cracker> will attempt cracking an <IsoRingedChain>. 
    """
    # NOTE: `suspected_isoring_to_sec_idn` can be incomplete w.r.t. `info`. In these cases, 
    # <BackgroundInfo> uses method<default_most_likely_Sec_index_for_IsoRing> to decide 
    # the best <HypStruct>. 
    def __init__(self,info:dict,suspected_isoring_to_sec_idn:dict,order_of_cracking:list): 
        assert BackgroundInfo.verify_valid_info(info) 

        q = set(info.keys()) 
        c = set() 
        for o in order_of_cracking: c |= o 
        assert q == c 

        self.info = info 
        self.suspected_isoring_to_sec_idn = suspected_isoring_to_sec_idn
        self.order_of_cracking = order_of_cracking
        return

    def hypothesis_for_IsoRingANDSec(self,i,s): 
        if not self.hypothesis_exists_for_IsoRingANDSec(i,s): 
            return None 
        return self.info[i][s] 

    def hypothesis_exists_for_IsoRingANDSec(self,i,s): 
        if i not in self.info: 
            return False 
        
        if s not in self.info[i]: 
            return False 
        return True 

    @staticmethod
    def verify_valid_info(info): 
        for k,v in info.items(): 
            if not type(k) == int: return False 
            for k2,v2 in v.items(): 
                if not type(k2) == int: return False 
                if not type(v2) == HypStruct: return False 
        return True 

    # TODO: test this. 
    @staticmethod
    def extract_from_IsoRingedChain(irc,prng,actual_sec_vec_ratio_range,dim_covered_ratio_range,\
        valid_bounds_ratio_range,prioritize_actual_Sec_ratio,shuffle_OOC_ratio,\
        suspected_isoring_to_sec_idn_error_ratio,valid_one_shot_kill_ratio_range=[0.,0.],hop_size_range=DEFAULT_HOP_SIZE_RANGE):
        assert type(irc) == IsoRingedChain 
        assert 0. <= actual_sec_vec_ratio_range[0] <= actual_sec_vec_ratio_range[1] <= 1.
        assert 0. <= dim_covered_ratio_range[0] <= dim_covered_ratio_range[1] <= 1.
        assert 0. <= valid_bounds_ratio_range[0] <= valid_bounds_ratio_range[1] <= 1.
        assert 0. <= prioritize_actual_Sec_ratio <= 1.0
        assert 0. <= shuffle_OOC_ratio <= 1.0
        assert 0. <= suspected_isoring_to_sec_idn_error_ratio <= 1.0

        def prg_(): return int(prng())

        prng_dec = prng_to_decimal_output(prng) 

        info = dict() 
        for idn_tag,ir in irc.ir_dict.items():
            # fetch all ratios and bools 
            if actual_sec_vec_ratio_range[0] == actual_sec_vec_ratio_range[1]: 
                actual_sec_vec_ratio = actual_sec_vec_ratio_range[0]
            else: 
                actual_sec_vec_ratio = modulo_in_range(prng_dec(),actual_sec_vec_ratio_range)

            if dim_covered_ratio_range[0] == dim_covered_ratio_range[1]: 
                ratio_of_dim_covered = dim_covered_ratio_range[0]
            else: 
                ratio_of_dim_covered = modulo_in_range(prng_dec(),dim_covered_ratio_range) 

            if valid_bounds_ratio_range[0] == valid_bounds_ratio_range[1]: 
                valid_bounds_ratio = valid_bounds_ratio_range[0] 
            else: 
                valid_bounds_ratio = modulo_in_range(prng_dec(),valid_bounds_ratio_range)

            if valid_one_shot_kill_ratio_range[0] == valid_one_shot_kill_ratio_range[1]: 
                valid_one_shot_kill_ratio = valid_one_shot_kill_ratio_range[0]
            else: 
                valid_one_shot_kill_ratio = modulo_in_range(prng_dec(),valid_one_shot_kill_ratio_range)

            prioritize_actual_Sec = prng_dec() <= prioritize_actual_Sec_ratio

            hdict = HypStruct.extract_from_IsoRing_into_HypStruct_dict(ir,prng,actual_sec_vec_ratio,\
                ratio_of_dim_covered,valid_bounds_ratio,prioritize_actual_Sec,valid_one_shot_kill_ratio,\
                hop_size_range)
            info[idn_tag] = hdict 

        # get the IsoRing identifiers for the IsoRing to Sec index error 
        i2s_error = int(ceil(suspected_isoring_to_sec_idn_error_ratio * len(info))) 
        L = sorted(info.keys())
        L = prg_seqsort(L,prg_)[:i2s_error] 
        S = dict() 
        for k in info.keys(): 
            ir = irc.fetch_IsoRing(k)
            
            # case: correct <Sec> index 
            actual = ir.actual_sec_index
            if k not in L: 
                S[k] = actual 
            else: 
                qs = [_ for _ in range(len(ir.sec_list)) if _ != actual]

                # case: no alternative Sec, use actual 
                if len(qs) == 0: 
                    S[k] = actual
                else: 
                    qi = prg_() % len(qs)
                    S[k] = qs[qi] 
               
        # get the OOC 
        num_swaps = int(ceil(shuffle_OOC_ratio * len(irc.ooc)))
        ooc = n_swaps_on_seq(deepcopy(irc.ooc),num_swaps,prg_)

        return BackgroundInfo(info,S,ooc) 

    def sec_index_for_IsoRing(self,ir_idn): 
        if ir_idn not in self.suspected_isoring_to_sec_idn: 
            return self.default_most_likely_Sec_index_for_IsoRing(ir_idn)
        return self.suspected_isoring_to_sec_idn[ir_idn]

    """
    chooses the <Sec> index with a <HypStruct> of highest probability marker. 
    """
    def default_most_likely_Sec_index_for_IsoRing(self,ir_idn): 
        if ir_idn not in self.info: return None 

        d = self.info[ir_idn] 
        if len(d) == 0: return None 

        secindex_pr = [] 
        for k,v in d.items(): 
            secindex_pr.append((k,v.probability_marker)) 
        
        secindex_pr = sorted(secindex_pr,key=lambda x:x[1],reverse=True)
        return secindex_pr[0][0]


from .crackling import * 
from morebs2.matrix_methods import is_number 

"""
Container to store cracked <Sec> instances from target <IsoRingedChain>.
"""
class CrackerSoln: 

    def __init__(self): 
        # IsoRing idn -> Sec index -> 
        #    (optima index, vector solution, ?does Crackling accept solution based on feedback Pr. value?) 
        self.D = dict() 

    def __len__(self): 
        n = 0 
        for v in self.D.values(): 
            n += len(v) 
        return n 

    def add_cracked_soln(self,cr:Crackling): 
        assert type(cr.cracked_soln) != type(None)
        assert type(cr.soln_pr) != None 
        if cr.target_ir not in self.D: 
            self.D[cr.target_ir] = dict() 
        
        assert cr.ir_sec_index not in self.D[cr.target_ir], "solution already acquired." 
        self.D[cr.target_ir][cr.ir_sec_index] = (cr.sec_opt_index,cr.cracked_soln,cr.soln_pr)
        return

    def fetch_soln(self,ir_idn,sec_index): 
        if ir_idn not in self.D: return None 
        if sec_index not in self.D[ir_idn]: return None  
        return self.D[ir_idn][sec_index] 

"""
<Cracker> is a structure that uses its given <BackgroundInfo> to conduct 
cracking on an <IsoRingedChain> in a brute-force environment. <Cracker> 
has the capacity to deploy at most `crackling_capacity` <Crackling>s at 
once. 

NOTE: 
<Cracker> may deploy more than 1 <Crackling> per <IsoRing>, although this 
is wasteful in an immobile brute-force environment such that <Crackling>s 
do not have to chase after their target <IsoRing>s. Mobile brute-force 
environments are not implemented in this program. Recall that the 
<BackgroundInfo> map is 
    <Isoring> identifier -> <Sec> index -> <HypStruct>. 
This means that every <Sec> for <IsoRing> is associated with at most 1 
<HypStruct>. Every <Crackling> that targets a <Sec> of an <IsoRing> will 
use the same <HypStruct>. 
"""
class Cracker: 

    def __init__(self,bi:BackgroundInfo,crackling_capacity:int,energy=float('inf'),verbose=False):
        assert type(bi) == BackgroundInfo 
        assert type(crackling_capacity) == int and crackling_capacity > 0 
        assert is_number(energy)
        assert type(verbose) == bool 

        self.bi = bi 
        self.crackling_capacity = crackling_capacity
        self.energy = energy 
        self.verbose = verbose 

        self.csoln = CrackerSoln() 
        # (IsoRing index, Sec index, optima index)
        self.failed_soln = [] 
        self.active_cracklings = [] 
        self.ooci = 0 
        self.target_ir_set = set()
        # successful cracks on element := (IsoRing idn, Sec idn)
        self.finished_target_list = [] 

        self.deployed_cracklings = False 
        # ?successfully finished or failed?
        self.halted = False  
        self.is_finished() 
        return

    def next_target_IsoRing_set(self):
        S = self.next_target_IsoRing_set_() 
        self.target_ir_set = S 
        if len(self.target_ir_set) == 0: return False 
        return True 

    def next_target_IsoRing_set_(self): 
        if self.verbose: print("-- Cracker cracks at index: {}".format(self.ooci))
        if len(self.bi.order_of_cracking) <= self.ooci: 
            return set() 
        return deepcopy(self.bi.order_of_cracking[self.ooci])

    def active_target_ir_size(self): 
        return len(self.target_ir_set) 

    """
    instantiate Cracklings based on info given by the two dict arguments. 

    return: 
    - ?deployment is successful? 
    """
    def deploy_cracklings(self,target_ir_to_isorepr_map:dict,target_ir_to_num_cracklings_map:dict): 
        self.deployed_cracklings = False 

        # case: halted from complete|incomplete cracking,no energy left. 
        if self.halted: return self.deployed_cracklings

        # case: no more target IsoRings 
        if len(self.target_ir_set) == 0: return self.deployed_cracklings

        # ensure crackling capacity can accomodate for number of cracklings
        #else: 
        target_irset = set(target_ir_to_num_cracklings_map.keys()) 
        assert set(target_ir_to_isorepr_map.keys()) == target_irset 
        assert target_irset.issubset(self.target_ir_set) 

        sum_wanted_cracklings = sum(target_ir_to_num_cracklings_map.values()) 

        # case: invalid Crackling allocation, error 
        if self.crackling_capacity - len(self.active_cracklings) < sum_wanted_cracklings: 
            return self.deployed_cracklings

        # allocate the Cracklings 
        for k,v in target_ir_to_num_cracklings_map.items(): 
            assert v > 0 
            for _ in range(v): 
                # case: no hypothesis for (IsoRing,Sec) exists, error 
                if not self.bi.hypothesis_exists_for_IsoRingANDSec(k,target_ir_to_isorepr_map[k]): 
                    self.active_cracklings.clear() 
                    return self.deployed_cracklings,0 

                hs = self.bi.info[k][target_ir_to_isorepr_map[k]] 
                c = Crackling(k,target_ir_to_isorepr_map[k],hs.opt_index)
                self.active_cracklings.append(c)

        self.deployed_cracklings = True 
        return self.deployed_cracklings 

    def manage_cracklings(self): 

        wanted_finishes,recracks = self.clear_finished_cracklings()

        # remove the finished IsoRing idns from `target_ir_set`
        self.target_ir_set -= wanted_finishes 

        # check if finished
        self.is_finished()
        return wanted_finishes,recracks 

    """
    return: 
    - idns of IsoRings where Cracklings cracked wanted Sec, 
      idns of IsoRings where recracking must occur for wanted Sec. 
    """
    def clear_finished_cracklings(self): 
        
        # set of IsoRing idns for Cracklings that terminated
        finished_ir_for_cracklings = set() 
        # set of IsoRing idns for Cracklings that terminated by 
        # successfully cracking the target Sec of the IsoRing.  
        wanted_finished_ir_for_cracklings = set() 

        l = 0 
        # iterate through and check each Crackling 
        while l < len(self.active_cracklings): 
            c = self.active_cracklings[l] 

            # case: extra Crackling for cracked (IsoRing,Sec)
            if c.target_ir in finished_ir_for_cracklings: 
                if self.verbose: print("-- Cracker clears extra crackling targeting IsoRing {}".format(c.target_ir))
                ##self.energy -= c.num_attempts 
                self.active_cracklings.pop(l) 
                continue 

            if c.terminated: 
                ##self.energy -= c.num_attempts 
                # case: solution acquired 
                if self.verbose: 
                    print("-- Crackling DONE.")
                    print(c) 

                if c.has_soln(): 
                    finished_ir_for_cracklings |= {c.target_ir}
                    self.csoln.add_cracked_soln(c) 
                    si = self.bi.sec_index_for_IsoRing(c.target_ir)
                    if si == c.ir_sec_index:
                        wanted_finished_ir_for_cracklings |= {c.target_ir}
                else:
                    if self.verbose: 
                        print("\t\t[!] no soln found")
                self.active_cracklings.pop(l) 
                continue 
            l += 1
        return wanted_finished_ir_for_cracklings,\
            finished_ir_for_cracklings - wanted_finished_ir_for_cracklings

    def soln_for_IsoRing(self,ir_idn):
        sec_index = self.bi.sec_index_for_IsoRing(ir_idn)
        return self.csoln.fetch_soln(ir_idn,sec_index) 

    def soln_synopsis(self): 
        print("------ CRACKER SOLUTION SYNOPSIS")
        for k in self.bi.info.keys(): 
            q = self.soln_for_IsoRing(k)
            print("[-] IsoRing {}: {}".format(k,q)) 
            print()
        

    """
    return:
    - ?failed at cracking any of the target IsoRings? 
    """
    def did_fail(self): 
        if not self.deployed_cracklings: 
            return False 

        if len(self.target_ir_set) > 0 and \
            len(self.active_cracklings) == 0: 
            self.halted = True 
            return True 
        return False 

    def is_finished(self): 
        if self.halted: return True 

        if self.energy <= 0: 
            self.halted = True 
            return True 

        return False 

    def increment_index(self): 
        self.ooci += 1



from .leak_functions import * 
from morebs2.matrix_methods import is_proper_bounds_vector

class HypStruct:

    def __init__(self,opt_index:int,suspected_subbound,hop_size,probability_marker):
        assert type(opt_index) == int 
        assert is_proper_bounds_vector(suspected_subbound) 
        assert type(hop_size) == int and hop_size > 1 
        assert type(probability_marker) in {type(None),float,np.float32,np.float64}, "got {}".format(type(probability_marker))
        if type(probability_marker) != type(None): 
            assert 0 <= probability_marker <= 1 

        self.opt_index = opt_index 
        self.suspected_subbound = suspected_subbound
        self.hop_size = hop_size
        self.probability_marker = probability_marker
        return

    def __str__(self): 
        s = "\t\tHypothesis" + "\n"
        s += "* opt index: " + str(self.opt_index) + "\n"
        s += "* suspected subbound: \n" + str(self.suspected_subbound) + "\n\n"
        s += "* hop size: " + str(self.hop_size) + "\n"
        s += "* probability: " + str(self.probability_marker) + "\n"
        return s 

    """
    - dict, sec index -> HypStruct 
    """
    @staticmethod 
    def extract_from_IsoRing_into_HypStruct_dict(ir:IsoRing,prng,actual_sec_vec_ratio=1.0,\
        ratio_of_dim_covered=1.0,valid_bounds_ratio=1.0,prioritize_actual_Sec:bool=True,\
        valid_one_shot_kill_ratio=0.0,hop_size_range=DEFAULT_HOP_SIZE_RANGE): 
        D = prng_leak_IsoRing_into_dict(ir,prng,actual_sec_vec_ratio,ratio_of_dim_covered,\
            valid_bounds_ratio,prioritize_actual_Sec,valid_one_shot_kill_ratio,hop_size_range=hop_size_range)  
        D2 = {} 

        for k,v in D.items(): 
            D2[k] = HypStruct(v[0],v[1],v[2],v[3]) 
        return D2 

    """
    method is port for receiving feedback information from <IsoRing> 
    """
    def register_pointANDpr(self,point,pr): 

        if not point_in_bounds(self.suspected_subbound,point): 
            return False 

        if pr != self.probability_marker: 
            return False 

        return True 


from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='isoring',
    version='0.1.8',
    description='A data security structure, IsoRing, and a brute-force environment.',
    long_description=readme,
    author='Richard Pham',
    author_email='phamrichard45@gmail.com',
    url='https://github.com/Changissnz/isoring',
    #license=LICENSE,
    packages=find_packages(exclude=('tests','docs'))
)



from .cracker import * 

def one_to_one_IsoRing2Crackling_map(ir_idn:set): 
    assert type(ir_idn) == set 
    return {i:1 for i in ir_idn} 

# used for long cracking sessions 
DEFAULT_MARKER = 10000 

"""
environment for a Cracker, given background information on an IsoRingedChain, to conduct brute-force guesses 
against it. 
"""
class BruteForceEnv:

    def __init__(self,crck:Cracker,irc:IsoRingedChain,prng=None,verbose=True): 
        assert type(crck) == Cracker and type(irc) == IsoRingedChain
        assert type(verbose) == bool 
        self.crck = crck 
        self.irc = irc 

        if type(prng) == type(None): 
            self.prng = default_std_Python_prng(output_range=[-10000,10000],rounding_depth=0) 
        else: 
            self.prng = prng 
        self.irc.prng = self.prng 
        self.verbose = verbose 

        self.cbridges = [] 
        self.num_iter = 0 
        return

    def __next__(self): 
        if self.verbose: 
            if self.num_iter % DEFAULT_MARKER == 0: 
                print("\t[!] iteration marker: ",self.num_iter)

        # case: finished 
        if self.is_finished(): return 

        # if no CBridges exist, declare them 
        if len(self.cbridges) == 0: 
            stat = self.first_cracking_bridges()
            if not stat: 
                return 
        
        self.run_cbridges() 
        self.num_iter += 1 
        return

    # TODO: finish 
    def is_finished(self): 
        self.crck.is_finished()
        if self.crck.halted: 
            return True 
        return False 

    def run_cbridges(self): 
        any_terminated = False 
        i = 0 
        while i < len(self.cbridges): 
            c = self.cbridges[i] 
            if c.terminated: 
                any_terminated = True 
                self.cbridges.pop(i)
                continue 
            r = next(c) 
            self.crck.energy -= r 
            i += 1 
            

        if any_terminated: 

            wanted_finishes,recracks = self.crck.manage_cracklings() 
            if self.verbose: 
                print("# BRIDGES: ",len(self.cbridges))
                print("\t- finished cracking on IsoRings\n{}".format(wanted_finishes)) 
                print("\t- recracking on IsoRings\n{}".format(recracks)) 
                print("=======================================================") 

            self.send_IsoRingedChain_info_on_cracked(wanted_finishes,recracks) 

            if len(recracks) > 0: 
                self.add_recracking_bridges(recracks)

        self.crck.did_fail() 

    def first_cracking_bridges(self): 
        stat = self.crck.next_target_IsoRing_set()

        if self.verbose: 
            print("[!] t={}, Adding initial cracking bridges to IsoRings\n{}\n".format(self.num_iter,self.crck.target_ir_set))

        # case: no new targets, terminated
        if not stat: 
            if self.verbose: print("[X] no new target IsoRings. DONE.")
            self.crck.halted = True 
            return stat  

        # case: wrong order, IsoRingedChain does not accept
        stat = self.irc.accept_cracker_targetset(self.crck.target_ir_set)
        if not stat: 
            if self.verbose: print("[X] wrong cracking order. DONE.")
            self.crck.halted = True 
            return stat 

        stat = self.initiate_Cracker_deployment(self.crck.target_ir_set)

        self.crck.increment_index() 
        return stat 

    def add_recracking_bridges(self,recrack_set):
        assert recrack_set.issubset(self.crck.target_ir_set) 
        return self.initiate_Cracker_deployment(recrack_set)

    def initiate_Cracker_deployment(self,ir_set): 

        # send Cracker info on IsoRing iso-repr state 
        ir_state = self.irc.repr_dict_for_IsoRings(ir_set)
        i2c_map = one_to_one_IsoRing2Crackling_map(ir_set)
            
        stat = self.crck.deploy_cracklings(ir_state,i2c_map)

        # case: Cracker is stuck. 
        if not stat: 
            self.crck.halted = True 
            return stat 

        # form the bridges 
        q = -sum(i2c_map.values())
        for cr in self.crck.active_cracklings[q:]: 
            hs = self.crck.bi.hypothesis_for_IsoRingANDSec(cr.target_ir,cr.ir_sec_index) 
            ir = self.irc.fetch_IsoRing(cr.target_ir)       
            cb = CBridge(cr,hs,ir) 
            self.cbridges.append(cb) 
        return True 

    def send_IsoRingedChain_info_on_cracked(self,wanted_finishes,recracks): 
        self.irc.register_cracked_IsoRings(wanted_finishes,recracks)
        return